from unittest import TestCase
from abl.util import Event, DispatcherObject, DONT_PROPAGATE


class Test_events(TestCase):

    def setUp(self):
        self.events = DispatcherObject()


    def test_first(self):
        empty_event = self.events.no_registered_event
        self.assert_(isinstance(empty_event, Event))


    def test_register(self):
        log = []
        def call_me(msg):
            log.append(msg)
        self.events.first_event.register(call_me)
        self.events.first_event.fire('my man')
        self.assertEqual(len(log), 1)
        self.assertEqual(log[0], 'my man')


    def test_detach(self):
        def nothing():pass
        self.events.some_event.register(nothing)
        self.assertEqual(len(self.events.some_event.callables), 1)
        self.events.some_event.detach(nothing)
        self.assertEqual(len(self.events.some_event.callables), 0)


    def test_detach_all(self):
        def nothing():pass
        self.events.some_event.register(nothing)
        self.events.some_event.register(nothing)
        self.events.some_event.register(nothing)
        self.assertEqual(len(self.events.some_event.callables), 3)
        self.events.some_event.detach_all()
        self.assertEqual(len(self.events.some_event.callables), 0)


    def test_register_classmethod(self):
        class MyClass(object):
            log = []

            @classmethod
            def run_me(cls, msg):
                cls.log.append(msg)

        self.events.run_me.register(MyClass.run_me)

        myobj = MyClass()
        self.events.run_me.fire('help!!!')
        self.assertEqual(myobj.log, ['help!!!'])


    def test_propagate(self):
        log = []
        def foo(msg):
            log.append(msg)

        def bar(msg):
            log.append(msg)
            return DONT_PROPAGATE

        def foobar(msg):
            log.append(msg)

        self.events.me.register(foo)
        self.events.me.register(bar)
        self.events.me.register(foobar)

        self.events.me.fire('help!!!')
        self.assertEqual(len(log), 2)


    def test_eventnames(self):
        self.events.foo.fire()
        self.events.bar.fire()
        self.assertEqual(len(self.events.eventnames()), 2)
        self.assertEqual(set(self.events.eventnames()), set(('foo', 'bar')))



from datetime import date


def disable(f):
    """Mark a test as disabled."""
    f.__test__ = False
    return f


def disable_if(predicate):
    """
    Disables a test if predicate yields True at parse time.

    @param predicate: a callable that gets passed the test as whole or the
        method under test and returns a bool
    @param reason: function that is responsible for the conditional, if nothing
        is passed, disable_if is assumed (used for disable_unless)
    @return: a decorator
    """
    def _decorator(f):
        if predicate():
            f.__test__ = False
        return f
    return _decorator


def disable_before(year, month=None, day=None):
    """
    Disables a test before the given day.

    Give year/month/day as int, and make sure not to pass in something like
    <09>, but pass <9> instead (<09> would be interpreted as being octal).

    @note: Please use this only when having a very good reason to disable
        a test before a certain date.

    @return: a decorator
    """
    if month is None and day is None:
        year, month, day = year.year, year.month, year.day
    def _decorator(f):
        if date.today() < date(year, month, day):
            f.__test__ = False
        return f
    return _decorator


def disable_unless(predicate):
    """
    Disables a test unless predicate yields True at parse time.

    @param predicate: a callable that gets passed the test
                      as whole or the method under test
                      and returns a bool
    @return: a decorator
    """
    def neg(*args):
        return not predicate(*args)
    return disable_if(neg)


def not_a_test(f):
    """Use this to mark methods that have test in their name but are not
    an actual test method."""
    f.__test__ = False
    f.__not_a_test__ = True
    return f

# not_a_test() is not a test
not_a_test = not_a_test(not_a_test)



import os
import logging
import errno

import platform
is_windows = False
if platform.system() == 'Windows':
    is_windows = True

if is_windows:
    import msvcrt #@UnresolvedImport
else:
    import fcntl


logger = logging.getLogger(__name__)

#-------------------------------------------------------------------------------

class LockFileCreationException(Exception):
    pass


#-------------------------------------------------------------------------------

class LockFileObtainException(Exception):
    pass



#-------------------------------------------------------------------------------

class LockFile(object):

    def __init__(self, name, fail_on_lock=False, cleanup=True):
        self.name = name
        self.cleanup = cleanup
        self.fail_on_lock = fail_on_lock
        self.fd = None
        self.file = None


    def __enter__(self):

        try:
            self.fd = os.open(self.name, os.O_WRONLY | os.O_CREAT | os.O_APPEND)
        except OSError, e:
            if e[0] == errno.ENOENT:
                raise LockFileCreationException()
            else:
                raise
        self.file = os.fdopen(self.fd, "w")
        if is_windows:
            lock_flags = msvcrt.LK_LOCK
        else:
            lock_flags = fcntl.LOCK_EX
        if self.fail_on_lock:
            if is_windows:
                lock_flags = msvcrt.LK_NBLCK
            else:
                lock_flags |= fcntl.LOCK_NB
        try:
            if is_windows:
                msvcrt.locking(self.file.fileno(), lock_flags, 1)
            else:
                fcntl.flock(self.file, lock_flags)
        except IOError, e:
            if e[0] == errno.EAGAIN:
                raise LockFileObtainException()
            raise


        return self.file


    def __exit__(self, unused_exc_type, unused_exc_val, unused_exc_tb):
        if not is_windows:
            fcntl.flock(self.file, fcntl.LOCK_UN)
        self.file.close()
        # we are told to cleanup after ourselves,
        # however it might be that another process
        # has done so - so we don't fail in that
        # case.
        if self.cleanup:
            try:
                os.remove(self.name)
            except OSError, e:
                if e[0] != errno.ENOENT:
                    raise




from __future__ import absolute_import

from .stream import (
    Stream,
    BufferedStream,
    )

from .memoization import (
    memoized,
    )

from .misc import (
    Bunch,
    NullHandler,
    SafeModifier,
    TeeBuffer,
    classproperty,
    fixpoint,
    partition,
    unicodify,
    with_,
    )

from .lockfile import (
    LockFile,
    LockFileCreationException,
    LockFileObtainException,
    )

from .dispatch import (
    Event,
    DispatcherObject,
    DONT_PROPAGATE,
    )



# -*- coding: utf-8 -*-

from __future__ import with_statement

from unittest import TestCase

from abl.util import (
    SafeModifier,
    classproperty,
    fixpoint,
    partition,
    unicodify,
    with_,
    )


class TestMisc(TestCase):


    def test_classproperty(self):

        value = 100
        class Foo(object):

            FOO = value

            @classproperty
            def test(cls):
                return cls.FOO

        assert Foo.test == value
        assert Foo().test == value


    def test_partition(self):
        self.assertEqual(partition(lambda i: i < 5, xrange(10)),
                         ([0, 1, 2, 3, 4], [5, 6, 7, 8, 9]))


class TestUnicodify(TestCase):

    def test_unicode_unity(self):
        utest = u"Hülle"
        assert unicodify(utest) is utest

    def test_utf8_to_unicode(self):
        teststring = 'H\xc3\xbclle'
        assert unicodify(teststring) == u"Hülle"

    def test_ignore_errors(self):
        teststring = 'H\xc3\xbclle'
        assert unicodify(teststring, codecs=()) == u"Hlle"


class FixpointTests(TestCase):

    def test_fixpoint(self):

        def f(x):
            if x == 10:
                return x
            return x + 1

        assert fixpoint(f, 1) == 10


class SafeModifierTests(TestCase):

    class Foo(object):
        pass


    def test_override_existing(self):
        foo = self.Foo()
        foo.bar = 42

        with SafeModifier(foo, 'bar', 17):
            self.assertEqual(foo.bar, 17)

        self.assertEqual(foo.bar, 42)


    def test_set_unset(self):
        foo = self.Foo()

        with SafeModifier(foo, 'bar', 17):
            self.assertEqual(foo.bar, 17)

        self.assertFalse(hasattr(foo, 'bar'))


class WithDecoratorTests(TestCase):

    class Foo(object):
        pass


    def test_decorate_function(self):

        foo = self.Foo()
        foo.bar = 42

        @with_(SafeModifier(foo, 'bar', 17))
        def decorated(x):
            self.assertEqual(foo.bar, 17)
            self.assertEqual(x, 1)

        decorated(1)
        self.assertEqual(foo.bar, 42)


    def test_decorate_method(self):

        foo = self.Foo()
        foo.bar = 42

        class MethodDecorationTest(object):
            @with_(SafeModifier(foo, 'bar', 13))
            def decorated(self, x):
                assert x == 1
                assert foo.bar == 13

        MethodDecorationTest().decorated(1)
        self.assertEqual(foo.bar, 42)


    def test_decorate_many(self):

        foo = self.Foo()
        foo.bar = 42
        foo.foo = 20

        class MethodDecorationTest(object):
            @with_(SafeModifier(foo, 'bar', 13), SafeModifier(foo, 'foo', 2))
            def decorated(self, x):
                assert x == 1
                assert foo.bar == 13
                assert foo.foo == 2

        MethodDecorationTest().decorated(1)
        self.assertEqual(foo.bar, 42)
        self.assertEqual(foo.foo, 20)




"""
The super-simple event system HOWTO
-----------------------------------

To register a function as an event handler::

    from ableton.dispatch import events

    @events.the_event.register
    def my_event_handler(arg1, arg2):
        do_some_stuff()


the_event is your event. You can use any name you want; if the event
is unknown it will be created automatically.

To fire an event and invoke all its handlers::

    from ableton.dispatch import events

    events.the_event.fire(arg1, arg2)

Note that you must fire the event using the same arguments that all
its handlers expect. If you don't, bad things will happen.

To prevent further handlers from processing an event::

    from ableton.dispatch import events, DONT_PROPOGATE

    @events.the_event.register
    def my_event_handler(arg1, arg2):
        do_some_stuff()
        return DONT_PROPOGATE

To detach your handler from an event::

    events.the_event.detach(my_event_handler)

"""

from collections import defaultdict

DONT_PROPAGATE = 'dont_propagate'


class Event(object):

    def __init__(self):
        self.callables = []


    def register(self, func):
        self.callables.append(func)
        return func


    def detach(self, func):
        if func in self.callables:
            self.callables.remove(func)


    def detach_all(self):
        self.callables = []


    def fire(self, *args, **kwargs):
        for func in self.callables:
            returnvalue = func(*args, **kwargs)
            if returnvalue is DONT_PROPAGATE:
                break



class DispatcherObject(object):

    def __init__(self):
        self.events = defaultdict(Event)


    def __getattr__(self, attr):
        return self.events[attr]


    def eventnames(self):
        return self.events.keys()


    def clear_all(self):
        self.events = defaultdict(Event)



from __future__ import with_statement

try:
    from multiprocessing import Process, Queue
except ImportError:
    print "This test needs the module multiprocessing."
    print "Please either use python >= 2.6 or install it."
else:
    from unittest import TestCase
    import tempfile
    import time

    from abl.util import LockFile, LockFileObtainException


    def locker(q, r):
        lock_file_name = q.get()
        with LockFile(lock_file_name):
            r.put("locked")
            while True:
                command = q.get()
                if command == "unlock":
                    break



    def failer(fname, r):
        try:
            with LockFile(fname, fail_on_lock=True):
                pass
        except LockFileObtainException:
            r.put("failed properly")
        except:
            r.put("something else went wrong")


    def waiter(fname, q, r):
        with LockFile(fname, fail_on_lock=True):
            r.put("got lock")
            # just wait until something happens
            q.get()


    class LockFileTests(TestCase):


        def test_lock_with_fail(self):
            q = Queue()
            r = Queue()
            locker_p = Process(target=locker, args=(q,r))
            locker_p.start()
            fname = tempfile.mktemp()
            q.put(fname)
            # wait until the locker really has the lock
            assert r.get() == "locked"
            # now start the failer
            failer_p = Process(target=failer, args=(fname, r))
            failer_p.start()

            failer_result = r.get()
            assert failer_result == "failed properly", failer_result
            # make the locker release it's lock
            q.put("unlock")


        def test_lock_with_wait(self):
            q = Queue()
            r = Queue()
            locker_p = Process(target=locker, args=(q,r))
            locker_p.start()
            fname = tempfile.mktemp()
            q.put(fname)
            # wait until the locker really has the lock
            assert r.get() == "locked"
            # now start the waiter
            waiter_p = Process(target=waiter, args=(fname, q, r))
            waiter_p.start()

            # now make the locker release the lock
            q.put("unlock")

            # this means the waiter should have it
            waiter_result = r.get()
            assert waiter_result == "got lock", waiter_result

            # and the failer should still fail, that poor fella
            failer_p = Process(target=failer, args=(fname, r))
            failer_p.start()

            failer_result = r.get()
            assert failer_result == "failed properly", failer_result

            # now let the waiter terminate
            q.put("whatever")


